package com.smfreports.activityreport;
                                                               
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.time.format.DateTimeFormatter;
import java.util.*;

import org.apache.commons.cli.*;

import static java.util.Comparator.*;

import com.blackhillsoftware.smf.SmfRecord;
import com.blackhillsoftware.smf.SmfRecordReader;
import com.blackhillsoftware.smf.smf30.Smf30Record;
import com.blackhillsoftware.util.DD;
     
import com.smfreports.common.*;

public class ActiveJobs                                                                            
{  
    public static void main(String[] args) throws IOException, ParseException                                   
    {   
        CommandLine cmd = initOptions(args);
        
        MatchList ignoreJobs = readIgnoreList(cmd, "ignorejobs");          
        MatchList ignoreUsers = readIgnoreList(cmd, "ignoreprograms");  

        Map<String, ActiveJobEntry> jobs = new HashMap<>();
        
        for (String inputFile : cmd.getArgList())
        {    
            System.out.format("Reading %s%n", inputFile);
            
            try (SmfRecordReader reader = SmfRecordReader.fromName(inputFile)
                    .include(30, 5))
            { 
                for (SmfRecord record : reader) 
                {                    
                    Smf30Record r30 = Smf30Record.from(record);
                    String jobname = r30.identificationSection().smf30jbn();
                    String user = r30.identificationSection().smf30uif();
                    if (r30.header().smf30wid().equals("JES2")
                            && !ignoreJobs.matchValue(jobname)
                            && !ignoreUsers.matchValue(user))
                    {
                        jobs.computeIfAbsent(jobname, key -> new ActiveJobEntry(jobname))
                            .add(r30);
                    }
                    
                }
            }
        }
               
        if (cmd.hasOption("csv"))
        {
            writeCsvReport(jobs);
        }
        else
        {
            writeReport(jobs);
        }
        
        // List ignored jobs and users for information and validation
        System.out.format("%n%s%n", "Ignored Jobs:");
        ignoreJobs.getMatches().stream()
            .sorted()
            .forEachOrdered(job -> System.out.format("%s%n", job));
        
        System.out.format("%n%s%n", "Ignored Users:");
        ignoreUsers.getMatches().stream()
            .sorted()
            .forEachOrdered(user -> System.out.format("%s%n", user)); 
    }

    private static void writeReport(Map<String, ActiveJobEntry> jobmap) 
    {
        System.out.format("%n%-8s  %-25s %6s  %8s  %-25s  %8s   %-25s%n", 
                "Jobname",
                "Last Run",
                "Count",
                "Max CPU",
                "Max CPU Submit Time",
                "Max EXCP", 
                "Max EXCP Submit Time");
        
        jobmap.values().stream()
                .sorted(comparing(ActiveJobEntry::getJobName))
                .forEachOrdered(entry ->
                {            
                    System.out.format("%-8s  %-25s %6d  %7.1fs  %-25s  %8d   %-25s%n", 
                            entry.getJobName(),
                            entry.getLastRunTime(),
                            entry.getCount(),
                            entry.getMaxCpu(),
                            entry.getMaxCpuSubmitTime(),
                            entry.getMaxExcp(), 
                            entry.getMaxExcpSubmitTime());
                });
    } 
    
    private static void writeCsvReport(Map<String, ActiveJobEntry> jobmap) 
    {
        
        DateTimeFormatter csvDateTime = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss"); 

        System.out.format("%s,%s,%s,%s,%s,%s,%s,%s,%s,%s%n", 
                "Jobname",
                "Last Run (ISO)",
                "Last Run (Excel)",
                "Count",
                "Max CPU",
                "Max CPU Submit Time (ISO)",
                "Max CPU Submit Time (Excel)",
                "Max EXCP", 
                "Max EXCP Submit Time (ISO)",
                "Max EXCP Submit Time (Excel)");
        
        jobmap.values().stream()
                .sorted(comparing(ActiveJobEntry::getJobName))
                .forEachOrdered(entry ->
                {            
                    System.out.format("%s,%s,%s,%d,%f,%s,%s,%d,%s,%s%n", 
                            entry.getJobName(),
                            entry.getLastRunTime(),
                            entry.getLastRunTime().format(csvDateTime),
                            entry.getCount(),
                            entry.getMaxCpu(),
                            entry.getMaxCpuSubmitTime(),
                            entry.getMaxCpuSubmitTime().format(csvDateTime),
                            entry.getMaxExcp(), 
                            entry.getMaxExcpSubmitTime(),
                            entry.getMaxExcpSubmitTime().format(csvDateTime));
                });
    } 
    
    private static MatchList readIgnoreList(CommandLine cmd, String optionName) throws IOException 
    {
        MatchList ignore = new MatchList();     
        String ignoreOption = cmd.getOptionValue(optionName); 
        if(ignoreOption != null)
        {
            ignore = ignoreOption.startsWith("//DD:") ?
                new MatchList(DD.readAllLines(ignoreOption)) :
                new MatchList(Files.readAllLines(new File(ignoreOption).toPath()));
        }
        return ignore;
    }
    
    private static void printUsage(Options options) 
    {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp( ActiveJobs.class.getName() + " [options] <input> ...", 
                System.lineSeparator() +
                "List jobs appearing in SMF type 30 data", 
                options, 
                System.lineSeparator() + 
                "<input> : File(s) containing SMF records. Binary data, RECFM=U or V[B] including RDW." +
                System.lineSeparator() +
                System.lineSeparator() +
                "<ignorejoblist>, <ignoreuserlist> and <input> can all be file names or DDNames in the format: //DD:DDNAME");
    }
    
    private static CommandLine initOptions(String[] args) throws ParseException 
    {
        Options options = new Options();
        options.addOption(
                Option.builder("h")
                    .longOpt("help")
                    .desc("print this message and exit")
                    .build());
        options.addOption("csv", "write output in CSV format");
        options.addOption( 
                Option.builder("ignorejobs")
                    .desc("a file or DD containing a list of jobs to exclude from the report" )
                    .hasArg()
                    .argName("ignorejoblist")
                    .build());
        options.addOption( 
                Option.builder("ignoreusers")
                    .desc("a file or DD containing a list of users to exclude from the report" )
                    .hasArg()
                    .argName("ignoreuserlist")
                    .build());
        
        CommandLineParser parser = new DefaultParser();
        CommandLine cmd = parser.parse(options, args);
        
        if (args.length == 0 || cmd.hasOption("help"))
        {
            printUsage(options);
            System.exit(0);
        }
        
        return cmd;
    }
    
}                                                                                               