package com.smfreports.activityreport;
                                                               
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.time.format.DateTimeFormatter;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.cli.*;

import static java.util.Comparator.*;

import com.blackhillsoftware.smf.SmfRecord;
import com.blackhillsoftware.smf.SmfRecordReader;
import com.blackhillsoftware.smf.smf30.Smf30Record;
import com.blackhillsoftware.util.DD;
            
import com.smfreports.common.*;

public class ActivePrograms                                                                            
{   
    public static void main(String[] args) throws IOException, ParseException                                   
    {   
        CommandLine cmd = initOptions(args);
               
        MatchList ignoreJobs = readIgnoreList(cmd, "ignorejobs");          
        MatchList ignorePrograms = readIgnoreList(cmd, "ignoreprograms");  
        
        Map<String, ActiveProgramEntry> programs = new HashMap<>();
        
        for (String inputFile : cmd.getArgList())
        {    
            System.out.format("Reading %s%n", inputFile);

            try (SmfRecordReader reader = SmfRecordReader.fromName(inputFile)
                    .include(30, 4))
            { 
                for (SmfRecord record : reader) 
                {                    
                    Smf30Record r30 = Smf30Record.from(record);
                    String jobname = r30.identificationSection().smf30jbn();
                    String program = r30.identificationSection().smf30pgm();
                    if (r30.header().smf30wid().equals("JES2")
                            && !ignoreJobs.matchValue(jobname)
                            && !ignorePrograms.matchValue(program))
                    {
                        programs.computeIfAbsent(program, key -> new ActiveProgramEntry(program))
                            .add(r30);
                    }          
                }
            }
        }
               
        if (cmd.hasOption("csv"))
        {
            writeCsvReport(programs);
        }
        else
        {
        	writeReport(programs);
        }        

        // List ignored jobs and users for information and validation
        System.out.format("%n%s%n", "Ignored Jobs:");
        ignoreJobs.getMatches().stream()
            .sorted()
            .forEachOrdered(job -> System.out.format("%s%n", job));
        
        System.out.format("%n%s%n", "Ignored Programs:");
        ignorePrograms.getMatches().stream()
            .sorted()
            .forEachOrdered(user -> System.out.format("%s%n", user));
        
    }

    private static void writeReport(Map<String, ActiveProgramEntry> programs) 
    {
        programs.entrySet().stream()
                .sorted(Map.Entry.comparingByKey())
                .forEachOrdered(entry ->
        {           
      		System.out.format("%n%-8s%n%n", entry.getKey());

            entry.getValue().getJobSteps().entrySet().stream()
                .sorted(Map.Entry.comparingByKey(
                        comparing(JobStep::getJobname)
                            .thenComparingInt(JobStep::getStepnumber)
                            .thenComparing(JobStep::getStepname)))
                .forEachOrdered(mapentry -> 
                {
                    System.out.format("         Job: %-8s   Step: %-3d     %-8s  Last Run: %-25s%n", 
                            mapentry.getKey().getJobname(), 
                            mapentry.getKey().getStepnumber(), 
                            mapentry.getKey().getStepname(),
                            mapentry.getValue());
                });
        });
    }
    
    private static void writeCsvReport(Map<String, ActiveProgramEntry> programs) 
    {
        DateTimeFormatter csvDateTime = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss"); 
    	
		System.out.format("%s,%s,%s,%s,%s,%s%n", 
				"Program Name",
				"Job Name",
				"Step",
				"Step Name",
				"Last Run (ISO)",
				"Last Run (Excel)"
				);        	
		
    	programs.entrySet().stream()
            .sorted(Map.Entry.comparingByKey())
            .forEachOrdered(entry ->              
	        {           
	            entry.getValue().getJobSteps().entrySet().stream()
	                .sorted(Map.Entry.comparingByKey(
	                        comparing(JobStep::getJobname)
	                            .thenComparingInt(JobStep::getStepnumber)
	                            .thenComparing(JobStep::getStepname)))
	                .forEachOrdered(mapentry -> 
	                {
	                    System.out.format("\"%s\",\"%s\",%d,\"%s\",%s,%s%n", 
	                    		entry.getKey(),
	                            mapentry.getKey().getJobname(), 
	                            mapentry.getKey().getStepnumber(), 
	                            mapentry.getKey().getStepname(),
	                            mapentry.getValue(),
	                            mapentry.getValue().format(csvDateTime));
	                });
	        });
    }
    
    private static MatchList readIgnoreList(CommandLine cmd, String optionName) throws IOException 
    {
        MatchList ignore = new MatchList();     
        String ignoreOption = cmd.getOptionValue(optionName); 
        if(ignoreOption != null)
        {
            ignore = ignoreOption.startsWith("//DD:") ?
                new MatchList(DD.readAllLines(ignoreOption)) :
                new MatchList(Files.readAllLines(new File(ignoreOption).toPath()));
        }
        return ignore;
    }
    
    private static void printUsage(Options options) 
    {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp( ActivePrograms.class.getName() + " [options] <input> ...", 
                System.lineSeparator() +
                "List programs appearing in SMF type 30 data", 
                options, 
                System.lineSeparator() + 
                "<input> : File(s) containing SMF records. Binary data, RECFM=U or V[B] including RDW." +
                System.lineSeparator() +
                System.lineSeparator() +
                "<ignorejoblist>, <ignoreprogramslist> and <input> can all be file names or DDNames in the format: //DD:DDNAME");
    }
    
    private static CommandLine initOptions(String[] args) throws ParseException 
    {
        Options options = new Options();
        options.addOption(
                Option.builder("h")
                    .longOpt("help")
                    .desc("print this message and exit")
                    .build());
        options.addOption("csv", "write output in CSV format");
        options.addOption( 
                Option.builder("ignorejobs")
                    .desc("a file or DD containing a list of jobs to exclude from the report" )
                    .hasArg()
                    .argName("ignorejoblist")
                    .build());
        options.addOption( 
                Option.builder("ignoreprograms")
                    .desc("a file or DD containing a list of programs to exclude from the report" )
                    .hasArg()
                    .argName("ignoreprogramlist")
                    .build());
        
        CommandLineParser parser = new DefaultParser();
        CommandLine cmd = parser.parse(options, args);
        
        if (args.length == 0 || cmd.hasOption("help"))
        {
            printUsage(options);
            System.exit(0);
        }
        
        return cmd;
    }
    
}                                                                                               