package com.smfreports.cics;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;

import org.apache.commons.cli.*;

import com.blackhillsoftware.smf.*;
import com.blackhillsoftware.smf.cics.*;
import com.blackhillsoftware.smf.cics.monitoring.*;
import com.blackhillsoftware.smf.cics.monitoring.fields.Field;

public class CicsProgramSummary 
{
    public static void main(String[] args) throws IOException, ParseException 
    {
        CommandLine cmd = initOptions(args);
        
        // create nested maps of Applid -> Program Name -> Transaction 
        Map<String, Map<String, Map<String, TransactionData>>> transactionsByApplid = new HashMap<>();

        int noDictionary = 0;
        int txCount = 0;

        // SmfRecordReader.fromName(...) accepts a filename, a DD name in the
        // format //DD:DDNAME or MVS dataset name in the form //'DATASET.NAME'
        
        for (String inputFile : cmd.getArgList())
        {    
            System.out.format("Reading %s%n", inputFile);
            
            try (SmfRecordReader reader = 
                    SmfRecordReader.fromName(inputFile)
                        .include(110, Smf110Record.SMFMNSTY)) 
            {     
                for (SmfRecord record : reader) 
                {
                    Smf110Record r110 = Smf110Record.from(record);
                    
                    if (r110.haveDictionary()) 
                    {
                        // Find the entry for this APPLID
                        Map<String, Map<String, TransactionData>> applidTransactions = 
                            transactionsByApplid
                                .computeIfAbsent(r110.mnProductSection().smfmnprn(), 
                                        transactions -> new HashMap<>());
    
                        for (PerformanceRecord perfRecord : r110.performanceRecords()) 
                        {
                            txCount++;
                            applidTransactions
                                // Lookup program name or create new entry
                                .computeIfAbsent(perfRecord.getField(Field.PGMNAME), 
                                        programName -> new HashMap<>())
                                // Lookup transaction name or create new entry
                                .computeIfAbsent(perfRecord.getField(Field.TRAN), 
                                        transactionData -> new TransactionData())
                                // Add transaction data
                                .add(perfRecord);
                        }
                    } 
                    else 
                    {
                        noDictionary++;
                    }
                }
            }
        }
        
        if (cmd.hasOption("csv"))
        {
            writeCsvReport(transactionsByApplid);
        }
        else
        {
        	writeReport(transactionsByApplid);
        }
        
        System.out.format(
                "%n%nTotal Transactions: %,d%n", 
                txCount);
               
        if (noDictionary > 0) 
        {
            System.out.format(
                    "%n%nSkipped %,d records because no applicable dictionary was found.", 
                    noDictionary);
        }
        
        // Write some compression statistics
        if (Smf110Record.getCompressedByteCount() > 0) 
        {
            System.out.format(
                    "%n%nCompressed bytes %,d, decompressed bytes %,d, compression %.1f%%.", 
                    Smf110Record.getCompressedByteCount(),
                    Smf110Record.getDecompressedByteCount(),
                    (double)(Smf110Record.getDecompressedByteCount() - Smf110Record.getCompressedByteCount()) 
                        / Smf110Record.getDecompressedByteCount() * 100);
        }
    }

    private static void writeReport(Map<String, Map<String, Map<String, TransactionData>>> transactionByApplid) 
    {        
        transactionByApplid.entrySet().stream()
            .sorted(Map.Entry.comparingByKey())
            .forEachOrdered(applidEntry -> 
            {
                // Headings
                System.out.format("%nAPPLID: %-8s", applidEntry.getKey());
                
                System.out.format("%n%n%-8s %-8s %15s %15s %15s %15s %15s %15s%n%n", 
                        "Program", 
                        "Tran", 
                        "Count", 
                        "Avg Elapsed", 
                        "CPU", 
                        "Avg CPU", 
                        "Avg Disp.", 
                        "Avg Disp Wait");
                
                // Get the program name entries for the applid
                applidEntry.getValue()
                    .entrySet().stream() // programs
                    .sorted(Map.Entry.comparingByKey())
                    .forEachOrdered(programEntry -> 
                    {
                        // Create a List so we can distinguish between
                        // the first and subsequent entries and write the program name
                        // in the first (or only) transaction line
                        
                        List<Entry<String, TransactionData>> transactions 
                            = new ArrayList<>(programEntry.getValue().entrySet());
                        transactions.sort(Map.Entry.comparingByKey());
                        
                        for (int i=0; i < transactions.size(); i++)
                        {
                            Entry<String, TransactionData> txEntry = transactions.get(i);
                            TransactionData txData = txEntry.getValue();
                            
                            System.out.format("%-8s %-8s %15d %15f %15f %15f %15f %15f%n",
                                    i == 0 ? programEntry.getKey() : "",
                                    txEntry.getKey(),
                                    txData.getCount(), 
                                    txData.getAvgElapsed(), 
                                    txData.getCpu(),
                                    txData.getAvgCpu(), 
                                    txData.getAvgDispatch(),
                                    txData.getAvgDispatchWait());
                            
                        }
                    });
            });

    }
    
    private static void writeCsvReport(Map<String, Map<String, Map<String, TransactionData>>> transactionByApplid) 
    {        
        // Headings               
        System.out.format("%s,%s,%s,%s,%s,%s,%s,%s,%s%n",
        		"Applid",
                "Program", 
                "Tran", 
                "Count", 
                "Avg Elapsed", 
                "CPU", 
                "Avg CPU", 
                "Avg Disp.", 
                "Avg Disp Wait");
        
        transactionByApplid.entrySet().stream()
            .sorted(Map.Entry.comparingByKey())
            .forEachOrdered(applidEntry -> 
            {
                // Get the program name entries for the applid
                applidEntry.getValue()
                    .entrySet().stream() // programs
                    .sorted(Map.Entry.comparingByKey())
                    .forEachOrdered(programEntry -> 
                    {
                    	programEntry.getValue().entrySet().stream()
                    		.sorted(Map.Entry.comparingByKey())
                    		.forEachOrdered(txEntry ->
							{
	                            TransactionData txData = txEntry.getValue();

	                            System.out.format("%s,%s,%s,%d,%f,%f,%f,%f,%f%n",
	                            		applidEntry.getKey(),
	                                    programEntry.getKey(),
	                                    txEntry.getKey(),
	                                    txData.getCount(), 
	                                    txData.getAvgElapsed(), 
	                                    txData.getCpu(),
	                                    txData.getAvgCpu(), 
	                                    txData.getAvgDispatch(),
	                                    txData.getAvgDispatchWait());
							
							});
                    });
            });
    }
    
    private static void printUsage(Options options) 
    {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp( CicsProgramSummary.class.getName() + " [options] <input> ...", 
                System.lineSeparator() +
                "Summarize CICS transaction data by program name and applid", 
                options, 
                System.lineSeparator() + 
                "<input> : File(s) containing SMF records. Binary data, RECFM=U or V[B] including RDW." +
                System.lineSeparator() +
                "CICS Dictionary records need to be read before the CICS transaction records." +
                System.lineSeparator() +
                "<input> can be file names or DDNames in the format: //DD:DDNAME");
    }
    
    private static CommandLine initOptions(String[] args) throws ParseException 
    {
        Options options = new Options();
        options.addOption(
                Option.builder("h")
                    .longOpt("help")
                    .desc("print this message and exit")
                    .build());
        options.addOption("csv", "write output in CSV format");
        
        CommandLineParser parser = new DefaultParser();
        CommandLine cmd = parser.parse(options, args);
        
        if (args.length == 0 || cmd.hasOption("help"))
        {
            printUsage(options);
            System.exit(0);
        }
        
        return cmd;
    }
}
