package com.smfreports.cics;

import java.io.*;
import java.util.*;

import org.apache.commons.cli.*;

import static java.util.Collections.reverseOrder;
import static java.util.Comparator.comparing;

import com.blackhillsoftware.smf.*;
import com.blackhillsoftware.smf.cics.*;
import com.blackhillsoftware.smf.cics.monitoring.*;
import com.blackhillsoftware.smf.cics.monitoring.fields.*;

public class CicsTransactionSummary 
{
    public static void main(String[] args) throws IOException, ParseException 
    {
        CommandLine cmd = initOptions(args);
        
        Map<String, Map<String, TransactionData>> transactionsByApplid = new HashMap<>();

        int noDictionary = 0;
        int txCount = 0;

        // SmfRecordReader.fromName(...) accepts a filename, a DD name in the
        // format //DD:DDNAME or MVS dataset name in the form //'DATASET.NAME'
        
        // Read multiple files or DDs if specified
        for (String inputFile : cmd.getArgList())
        {    
            System.out.format("Reading %s%n", inputFile);
    
            try (SmfRecordReader reader = 
                    SmfRecordReader.fromName(inputFile)
                        .include(110, Smf110Record.SMFMNSTY)) 
            {     
                for (SmfRecord record : reader) 
                {
                    Smf110Record r110 = Smf110Record.from(record);
                    
                    if (r110.haveDictionary()) 
                    {
                        Map<String, TransactionData> applidTransactions = 
                            transactionsByApplid.computeIfAbsent(
                                r110.mnProductSection().smfmnprn(), 
                                transactions -> new HashMap<>());
    
                        for (PerformanceRecord perfRecord : r110.performanceRecords()) 
                        {
                            txCount++;
                            applidTransactions.computeIfAbsent(
                                    perfRecord.getField(Field.TRAN), 
                                    x -> new TransactionData()).add(perfRecord);
                        }
                    } 
                    else 
                    {
                        noDictionary++;
                    }
                }
            }
        }
        
        if (cmd.hasOption("csv"))
        {
            writeCsvReport(transactionsByApplid);
        }
        else
        {
        	writeReport(transactionsByApplid);
        }
        
        System.out.format(
                "%n%nTotal Transactions: %,d%n", 
                txCount);
               
        if (noDictionary > 0) 
        {
            System.out.format(
                    "%n%nSkipped %,d records because no applicable dictionary was found.", 
                    noDictionary);
        }
        
        if (Smf110Record.getCompressedByteCount() > 0) 
        {
            System.out.format(
                    "%n%nCompressed bytes %,d, decompressed bytes %,d, compression %.1f%%.", 
                    Smf110Record.getCompressedByteCount(),
                    Smf110Record.getDecompressedByteCount(),
                    (double)(Smf110Record.getDecompressedByteCount() - Smf110Record.getCompressedByteCount()) / Smf110Record.getDecompressedByteCount() * 100);
        }
    }

    private static void writeReport(Map<String, Map<String, TransactionData>> transactions) 
    {
        transactions.entrySet().stream()
            .sorted(Map.Entry.comparingByKey())
            .forEachOrdered(applidEntry -> 
            {
                // Headings
                System.out.format("%n%-8s", applidEntry.getKey());
    
                System.out.format("%n%-4s %15s %15s %15s %15s %15s %15s%n%n", 
                        "Name", 
                        "Count", 
                        "Avg Elapsed", 
                        "CPU", 
                        "Avg CPU", 
                        "Avg Disp.", 
                        "Avg Disp Wait");
    
                applidEntry.getValue()
                    .entrySet().stream() // transactions
                    .sorted(Map.Entry.comparingByValue(
                            comparing(TransactionData::getCpu, reverseOrder())
                            .thenComparing(TransactionData::getCount, reverseOrder())))
                    .forEachOrdered(transactionEntry -> 
                    {
                        TransactionData txData = transactionEntry.getValue();
                        
                        // write detail line
                        System.out.format("%-4s %15d %15f %15f %15f %15f %15f%n", 
                                transactionEntry.getKey(),
                                txData.getCount(), 
                                txData.getAvgElapsed(), 
                                txData.getCpu(),
                                txData.getAvgCpu(), 
                                txData.getAvgDispatch(),
                                txData.getAvgDispatchWait());

                    });
            });

    }
    
    private static void writeCsvReport(Map<String, Map<String, TransactionData>> transactions) 
    {
        // Headings
    	System.out.format("%s,%s,%s,%s,%s,%s,%s,%s%n",
        		"Applid",
                "Name", 
                "Count", 
                "Avg Elapsed", 
                "CPU", 
                "Avg CPU", 
                "Avg Disp.", 
                "Avg Disp Wait");
        
        transactions.entrySet().stream()
            .sorted(Map.Entry.comparingByKey())
            .forEachOrdered(applidEntry -> 
            {
                applidEntry.getValue()
                    .entrySet().stream() // transactions
                    .sorted(Map.Entry.comparingByKey())
                    .forEachOrdered(transactionEntry -> 
                    {
                        TransactionData txData = transactionEntry.getValue();
                        
                        System.out.format("%s,%s,%d,%f,%f,%f,%f,%f%n", 
                        		applidEntry.getKey(),
                                transactionEntry.getKey(),
                                txData.getCount(), 
                                txData.getAvgElapsed(), 
                                txData.getCpu(),
                                txData.getAvgCpu(), 
                                txData.getAvgDispatch(),
                                txData.getAvgDispatchWait());

                    });
            });

    }
    
    private static void printUsage(Options options) 
    {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp( CicsTransactionSummary.class.getName() + " [options] <input> ...", 
                System.lineSeparator() +
                "Summarize CICS transaction data by transaction name and applid", 
                options, 
                System.lineSeparator() + 
                "<input> : File(s) containing SMF records. Binary data, RECFM=U or V[B] including RDW." +
                System.lineSeparator() +
                "CICS Dictionary records need to be read before the CICS transaction records." +
                System.lineSeparator() +
                "<input> can be file names or DDNames in the format: //DD:DDNAME");
    }
    
    private static CommandLine initOptions(String[] args) throws ParseException 
    {
        Options options = new Options();
        options.addOption(
                Option.builder("h")
                    .longOpt("help")
                    .desc("print this message and exit")
                    .build());
        options.addOption("csv", "write output in CSV format");
        
        CommandLineParser parser = new DefaultParser();
        CommandLine cmd = parser.parse(options, args);
        
        if (args.length == 0 || cmd.hasOption("help"))
        {
            printUsage(options);
            System.exit(0);
        }
        
        return cmd;
    }
    
}
