package com.smfreports.common;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * Test strings against a list of patterns.
 * To avoid the complexity of regular expressions a
 * simplified syntax is used and converted to a regular expression.
 * Existing dots "." i.e. separators between dataset qualifiers are
 * escaped.
 * An asterisk "*" matches 1 or more of any characters.
 * A question mark "?" matches a single character.
 * Regular expression elements that don't use ".", "*" or "?" can
 * be used.    
 *
 */
public class MatchList 
{
	private List<Pattern> patterns = new ArrayList<>();
	
	// Testing names against each pattern is potentially
	// expensive as the list gets longer, so store
	// the result for each name seen.
	private Map<String, Boolean> cachedResults = new HashMap<>();
	
	boolean empty = true;
	
	public MatchList()
    {
    }
	
	public MatchList(List<String> list)
	{
		if (list.size() > 0)
		{
		    empty = false;
		}
		for (String entry : list)
		{
			patterns.add(buildPattern(entry));
		}

	}	
	
	/**
	 * Test whether a value matches anything in the list, 
	 * keeping a cache of results to avoid
	 * checking the whole list for values that have been seen before
	 * 
	 * @param value The value to check
	 * @return true if the value matches
	 */
	public boolean matchValue(String value) 
	{
	    if (empty) return false; // shortcut - nothing matches an empty list
		Boolean result = cachedResults.get(value);
		if (result != null)
		{
			return result;
		}
		else
		{
			for (Pattern pattern : patterns)
			{
				if (pattern.matcher(value).matches())
				{
					cachedResults.put(value, Boolean.TRUE);
					return true;
				}
			}
			// no patterns match
			cachedResults.put(value, Boolean.FALSE);
			return false;
		}
	}
	
	private static Pattern buildPattern(String patternString) {
		patternString = patternString.replace(".", "\\.");
		patternString = patternString.replace("?", ".");
		patternString = patternString.replace("*", ".+");
		return Pattern.compile("^" + patternString + "$");
	}
	
	public List<String> getMatches()
	{
		return cachedResults.entrySet().stream()
				.filter(entry -> entry.getValue() == Boolean.TRUE)
				.map(entry -> entry.getKey())
				.collect(Collectors.toList());
	}
}
